functor LexerFn(structure Tokens : Parser_TOKENS)  = struct

    structure yyInput : sig

        type stream
    val mkStream : (int -> string) -> stream
    val fromStream : TextIO.StreamIO.instream -> stream
    val getc : stream -> (Char.char * stream) option
    val getpos : stream -> int
    val getlineNo : stream -> int
    val subtract : stream * stream -> string
    val eof : stream -> bool
    val lastWasNL : stream -> bool

      end = struct

        structure TIO = TextIO
        structure TSIO = TIO.StreamIO
    structure TPIO = TextPrimIO

        datatype stream = Stream of {
            strm : TSIO.instream,
        id : int,  (* track which streams originated 
            * from the same stream *)
        pos : int,
        lineNo : int,
        lastWasNL : bool
          }

    local
      val next = ref 0
    in
    fun nextId() = !next before (next := !next + 1)
    end

    val initPos = 2 (* ml-lex bug compatibility *)

    fun mkStream inputN = let
              val strm = TSIO.mkInstream 
               (TPIO.RD {
                    name = "lexgen",
                chunkSize = 4096,
                readVec = SOME inputN,
                readArr = NONE,
                readVecNB = NONE,
                readArrNB = NONE,
                block = NONE,
                canInput = NONE,
                avail = (fn () => NONE),
                getPos = NONE,
                setPos = NONE,
                endPos = NONE,
                verifyPos = NONE,
                close = (fn () => ()),
                ioDesc = NONE
                  }, "")
          in 
        Stream {strm = strm, id = nextId(), pos = initPos, lineNo = 1,
            lastWasNL = true}
          end

    fun fromStream strm = Stream {
        strm = strm, id = nextId(), pos = initPos, lineNo = 1, lastWasNL = true
          }

    fun getc (Stream {strm, pos, id, lineNo, ...}) = (case TSIO.input1 strm
              of NONE => NONE
           | SOME (c, strm') => 
           SOME (c, Stream {
                    strm = strm', 
                pos = pos+1, 
                id = id,
                lineNo = lineNo + 
                     (if c = #"\n" then 1 else 0),
                lastWasNL = (c = #"\n")
                  })
         (* end case*))

    fun getpos (Stream {pos, ...}) = pos

    fun getlineNo (Stream {lineNo, ...}) = lineNo

    fun subtract (new, old) = let
          val Stream {strm = strm, pos = oldPos, id = oldId, ...} = old
          val Stream {pos = newPos, id = newId, ...} = new
              val (diff, _) = if newId = oldId andalso newPos >= oldPos
                  then TSIO.inputN (strm, newPos - oldPos)
                  else raise Fail 
                "BUG: yyInput: attempted to subtract incompatible streams"
          in 
        diff 
          end

    fun eof (Stream {strm, ...}) = TSIO.endOfStream strm

    fun lastWasNL (Stream {lastWasNL, ...}) = lastWasNL

      end

    datatype yystart_state = 
COMMENT | LCOMMENT | INITIAL
    structure UserDeclarations = 
      struct

open Tokens

(* Types to match structure LEXER.UserDeclaration *)
type ('a,'b) token = ('a,'b) Tokens.token
type pos           = EL.pos
type svalue        = Tokens.svalue
type lexresult     = (svalue, pos) token

(* Handling lines and nested comments *)
val line = ref 1
val lines = ref [0]
val nesting = ref 0
val comments = ref [] : int list ref

fun newline yypos =
    (line := !line+1; lines := (yypos+1) :: !lines)
fun nest yypos =
    (nesting := !nesting+1; comments := yypos :: !comments)
fun unnest() =
    (nesting := !nesting-1; comments := List.tl(!comments))

fun lineColumn pos = lineColumn'(pos, !line, !lines)
and lineColumn'(pos, line, []) = (line, pos)
  | lineColumn'(pos, line, p::ps) =
    if pos >= p then (line, pos-p) else lineColumn'(pos, line-1, ps)

(*
val newline = fn pos =>
    (
        newline pos;
        print("[newline "^Int.toString(pos+1)^"]")
    )
val lineColumn = fn pos =>
    let
        val (l, c) = lineColumn pos
    in
        print("[token "^ Int.toString pos ^" -> "^Int.toString l^"."^Int.toString c^"]");
        (l, c)
    end
*)

fun toRegion(yypos, yytext) =
    let
        val left = yypos
        val right = left + String.size yytext
    in
        (lineColumn left, lineColumn right)
    end

fun eof() =
    let
        val lastLine = lineColumn(List.hd(!lines))  (* we ensure that last line is always empty *)
    in
        if !nesting = 0
        then Tokens.EOF(lastLine, lastLine)
        else raise EL.Error({l=lineColumn(List.hd(!comments)), r=lastLine}, "unclosed comment")
    end

(* Some helpers to create tokens *)
open Tokens

fun token(TOKEN, yypos, yytext) =
    TOKEN(toRegion(yypos, yytext))

fun tokenOf(TOKEN, toVal, yypos, yytext) =
    let
        val (l, r) = toRegion(yypos, yytext)
    in
        TOKEN(toVal yytext, l, r)
    end

fun error(yypos, yytext, s) =
    let
        val (l, r) = toRegion(yypos, yytext)
    in
        raise EL.Error({l=l, r=r}, s)
    end

fun invalid(yypos, yytext) =
    let
        val s = "invalid character `" ^ String.toString yytext ^ "'"
    in
        error(yypos, yytext, s)
    end

(* Convert identifiers *)
fun toId s = s

(* Convert constants *)
fun toInt s     = valOf(Int.fromString s)
fun toHexInt s  = valOf(StringCvt.scanString (Int.scan StringCvt.HEX) (String.substring(s, 2, String.size s-2)))
fun toReal s    = valOf(Real.fromString s)
fun toString s  = valOf(String.fromString(String.substring(s, 1, String.size s-2)))
fun toChar s    = valOf(Char.fromString(String.substring(s, 2, String.size s-3)))



      end

    datatype yymatch 
      = yyNO_MATCH
      | yyMATCH of yyInput.stream * action * yymatch
    withtype action = yyInput.stream * yymatch -> UserDeclarations.lexresult

    local

    val yytable = 
#[
([(#"\^@",#"\t",3),
(#"\v",#"'",3),
(#")",#")",3),
(#"+",#"\255",3),
(#"\n",#"\n",4),
(#"(",#"(",5),
(#"*",#"*",6)], []), ([(#"\^@",#"\t",10),
(#"\v",#"\255",10),
(#"\n",#"\n",11)], []), ([(#"\^@",#"\b",12),
(#"\^N",#"\^_",12),
(#"\127",#"\255",12),
(#"\t",#"\t",13),
(#"\v",#"\r",13),
(#" ",#" ",13),
(#"\n",#"\n",14),
(#"!",#"!",15),
(#"\"",#"\"",16),
(#"#",#"#",17),
(#"$",#"&",18),
(#"*",#"*",18),
(#"/",#"/",18),
(#">",#"@",18),
(#"\\",#"\\",18),
(#"^",#"^",18),
(#"`",#"`",18),
(#"'",#"'",19),
(#"(",#"(",20),
(#")",#")",21),
(#"+",#"+",22),
(#",",#",",23),
(#"-",#"-",24),
(#".",#".",25),
(#"0",#"0",26),
(#"1",#"9",27),
(#":",#":",28),
(#";",#";",29),
(#"<",#"<",30),
(#"=",#"=",31),
(#"A",#"Z",32),
(#"a",#"a",32),
(#"c",#"c",32),
(#"g",#"h",32),
(#"j",#"k",32),
(#"q",#"q",32),
(#"x",#"z",32),
(#"[",#"[",33),
(#"]",#"]",34),
(#"_",#"_",35),
(#"b",#"b",36),
(#"d",#"d",37),
(#"e",#"e",38),
(#"f",#"f",39),
(#"i",#"i",40),
(#"l",#"l",41),
(#"m",#"m",42),
(#"n",#"n",43),
(#"o",#"o",44),
(#"p",#"p",45),
(#"r",#"r",46),
(#"s",#"s",47),
(#"t",#"t",48),
(#"u",#"u",49),
(#"v",#"v",50),
(#"w",#"w",51),
(#"{",#"{",52),
(#"|",#"|",53),
(#"}",#"}",54),
(#"~",#"~",55)], []), ([], [73]), ([], [74]), ([(#"*",#"*",8)], [73]), ([(#")",#")",7)], [73]), ([], [72]), ([(#")",#")",9)], [71]), ([], [70]), ([], [68]), ([], [69]), ([], [76]), ([(#"\t",#"\t",199),
(#"\v",#"\r",199),
(#" ",#" ",199)], [0, 76]), ([], [1]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56)], [2, 65, 76]), ([(#" ",#"!",188),
(#"#",#"[",188),
(#"]",#"~",188),
(#"\128",#"\255",188),
(#"\"",#"\"",189),
(#"\\",#"\\",190)], [75, 76]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56),
(#"\"",#"\"",174)], [6, 65, 76]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56)], [65, 76]), ([(#"'",#"'",173),
(#"0",#"9",173),
(#"A",#"Z",173),
(#"_",#"_",173),
(#"a",#"z",173)], [63, 76]), ([(#"*",#"*",171)], [7, 76]), ([], [8, 76]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"*",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56),
(#"+",#"+",170)], [3, 65, 76]), ([], [9, 76]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"=",56),
(#"?",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56),
(#">",#">",169)], [5, 65, 76]), ([], [11, 76]), ([(#".",#".",59),
(#"0",#"9",58),
(#"E",#"E",60),
(#"e",#"e",60),
(#"x",#"x",64)], [58, 76]), ([(#".",#".",59),
(#"0",#"9",58),
(#"E",#"E",60),
(#"e",#"e",60)], [58, 76]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"=",56),
(#"?",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56),
(#">",#">",168)], [12, 65, 76]), ([], [14, 76]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56)], [15, 65, 76]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"<",56),
(#"?",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56),
(#"=",#"=",166),
(#">",#">",167)], [16, 65, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [64, 76]), ([], [19, 76]), ([], [20, 76]), ([], [21, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"n",66),
(#"p",#"z",66),
(#"o",#"o",163)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"b",#"n",66),
(#"p",#"z",66),
(#"a",#"a",159),
(#"o",#"o",160)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"k",66),
(#"m",#"m",66),
(#"o",#"w",66),
(#"y",#"z",66),
(#"l",#"l",149),
(#"n",#"n",150),
(#"x",#"x",151)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"b",#"m",66),
(#"p",#"t",66),
(#"v",#"z",66),
(#"a",#"a",137),
(#"n",#"n",138),
(#"o",#"o",139),
(#"u",#"u",140)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"e",66),
(#"g",#"l",66),
(#"o",#"z",66),
(#"f",#"f",129),
(#"m",#"m",130),
(#"n",#"n",131)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"h",66),
(#"j",#"z",66),
(#"e",#"e",124),
(#"i",#"i",125)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"n",66),
(#"p",#"z",66),
(#"o",#"o",119)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"z",66),
(#"e",#"e",117)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"e",66),
(#"g",#"o",66),
(#"q",#"t",66),
(#"v",#"z",66),
(#"f",#"f",111),
(#"p",#"p",112),
(#"u",#"u",113)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"q",66),
(#"s",#"z",66),
(#"r",#"r",107)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"z",66),
(#"e",#"e",105)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"h",66),
(#"j",#"s",66),
(#"u",#"z",66),
(#"e",#"e",88),
(#"i",#"i",89),
(#"t",#"t",90)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"g",66),
(#"i",#"q",66),
(#"s",#"x",66),
(#"z",#"z",66),
(#"h",#"h",79),
(#"r",#"r",80),
(#"y",#"y",81)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"m",66),
(#"o",#"z",66),
(#"n",#"n",76)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"b",#"z",66),
(#"a",#"a",74)], [64, 76]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"g",66),
(#"j",#"z",66),
(#"h",#"h",67),
(#"i",#"i",68)], [64, 76]), ([], [22, 76]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56)], [23, 65, 76]), ([], [24, 76]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56),
(#"0",#"0",57),
(#"1",#"9",58)], [65, 76]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56)], [65]), ([(#".",#".",59),
(#"0",#"9",58),
(#"E",#"E",60),
(#"e",#"e",60),
(#"x",#"x",64)], [58]), ([(#".",#".",59),
(#"0",#"9",58),
(#"E",#"E",60),
(#"e",#"e",60)], [58]), ([(#"0",#"9",63)], []), ([(#"0",#"9",61),
(#"~",#"~",62)], []), ([(#"0",#"9",61)], [60]), ([(#"0",#"9",61)], []), ([(#"0",#"9",63),
(#"E",#"E",60),
(#"e",#"e",60)], [60]), ([(#"0",#"9",65),
(#"A",#"F",65),
(#"a",#"f",65)], []), ([(#"0",#"9",65),
(#"A",#"F",65),
(#"a",#"f",65)], [59]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"z",66),
(#"e",#"e",71)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"s",66),
(#"u",#"z",66),
(#"t",#"t",69)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"g",66),
(#"i",#"z",66),
(#"h",#"h",70)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [57, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"q",66),
(#"s",#"z",66),
(#"r",#"r",72)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"z",66),
(#"e",#"e",73)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [56, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"k",66),
(#"m",#"z",66),
(#"l",#"l",75)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [55, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"h",66),
(#"j",#"z",66),
(#"i",#"i",77)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"s",66),
(#"u",#"z",66),
(#"t",#"t",78)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [54, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"z",66),
(#"e",#"e",86)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"t",66),
(#"v",#"z",66),
(#"u",#"u",84)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"o",66),
(#"q",#"z",66),
(#"p",#"p",82)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"z",66),
(#"e",#"e",83)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [53, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"z",66),
(#"e",#"e",85)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [52, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"m",66),
(#"o",#"z",66),
(#"n",#"n",87)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [51, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"b",#"z",66),
(#"a",#"a",102)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"f",66),
(#"h",#"z",66),
(#"g",#"g",95)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"q",66),
(#"s",#"z",66),
(#"r",#"r",91)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"h",66),
(#"j",#"z",66),
(#"i",#"i",92)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"m",66),
(#"o",#"z",66),
(#"n",#"n",93)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"f",66),
(#"h",#"z",66),
(#"g",#"g",94)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [50, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"m",66),
(#"o",#"z",66),
(#"n",#"n",96)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"b",#"z",66),
(#"a",#"a",97)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"s",66),
(#"u",#"z",66),
(#"t",#"t",98)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"t",66),
(#"v",#"z",66),
(#"u",#"u",99)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"q",66),
(#"s",#"z",66),
(#"r",#"r",100)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"z",66),
(#"e",#"e",101)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [49, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"k",66),
(#"m",#"z",66),
(#"l",#"l",103)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"r",66),
(#"t",#"z",66),
(#"s",#"s",104)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [48, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"b",66),
(#"d",#"z",66),
(#"c",#"c",106)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [47, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"h",66),
(#"j",#"z",66),
(#"i",#"i",108)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"m",66),
(#"o",#"z",66),
(#"n",#"n",109)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"s",66),
(#"u",#"z",66),
(#"t",#"t",110)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [46, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [43, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"z",66),
(#"e",#"e",115)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"s",66),
(#"u",#"z",66),
(#"t",#"t",114)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [45, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"m",66),
(#"o",#"z",66),
(#"n",#"n",116)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [44, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"v",66),
(#"x",#"z",66),
(#"w",#"w",118)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [42, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"c",66),
(#"e",#"z",66),
(#"d",#"d",120)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"t",66),
(#"v",#"z",66),
(#"u",#"u",121)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"k",66),
(#"m",#"z",66),
(#"l",#"l",122)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"z",66),
(#"e",#"e",123)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [41, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"s",66),
(#"u",#"z",66),
(#"t",#"t",128)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"m",66),
(#"o",#"z",66),
(#"n",#"n",126)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"j",66),
(#"l",#"z",66),
(#"k",#"k",127)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [40, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [39, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [35, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"o",66),
(#"q",#"z",66),
(#"p",#"p",133)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"s",66),
(#"u",#"z",66),
(#"t",#"t",132)], [37, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [38, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"n",66),
(#"p",#"z",66),
(#"o",#"o",134)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"q",66),
(#"s",#"z",66),
(#"r",#"r",135)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"s",66),
(#"u",#"z",66),
(#"t",#"t",136)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [36, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"k",66),
(#"m",#"z",66),
(#"l",#"l",146)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [31, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"q",66),
(#"s",#"z",66),
(#"r",#"r",142)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"m",66),
(#"o",#"z",66),
(#"n",#"n",141)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [34, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"b",#"z",66),
(#"a",#"a",143)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"k",66),
(#"m",#"z",66),
(#"l",#"l",144)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"k",66),
(#"m",#"z",66),
(#"l",#"l",145)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [33, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"r",66),
(#"t",#"z",66),
(#"s",#"s",147)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"z",66),
(#"e",#"e",148)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [32, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"r",66),
(#"t",#"z",66),
(#"s",#"s",157)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"c",66),
(#"e",#"z",66),
(#"d",#"d",156)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"o",66),
(#"q",#"z",66),
(#"p",#"p",152)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"n",66),
(#"p",#"z",66),
(#"o",#"o",153)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"q",66),
(#"s",#"z",66),
(#"r",#"r",154)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"s",66),
(#"u",#"z",66),
(#"t",#"t",155)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [30, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [29, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"d",66),
(#"f",#"z",66),
(#"e",#"e",158)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [28, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"s",66),
(#"u",#"z",66),
(#"t",#"t",161)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [27, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"b",#"z",66),
(#"a",#"a",162)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [26, 64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"n",66),
(#"p",#"z",66),
(#"o",#"o",164)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"k",66),
(#"m",#"z",66),
(#"l",#"l",165)], [64]), ([(#"'",#"'",66),
(#"0",#"9",66),
(#"A",#"Z",66),
(#"_",#"_",66),
(#"a",#"z",66)], [25, 64]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56)], [17, 65]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56)], [18, 65]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56)], [13, 65]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56)], [10, 65]), ([(#"!",#"!",56),
(#"#",#"&",56),
(#"*",#"+",56),
(#"-",#"-",56),
(#"/",#"/",56),
(#":",#":",56),
(#"<",#"@",56),
(#"\\",#"\\",56),
(#"^",#"^",56),
(#"`",#"`",56),
(#"|",#"|",56),
(#"~",#"~",56)], [4, 65]), ([(#")",#")",172)], [67]), ([], [66]), ([(#"'",#"'",173),
(#"0",#"9",173),
(#"A",#"Z",173),
(#"_",#"_",173),
(#"a",#"z",173)], [63]), ([(#" ",#"!",175),
(#"#",#"[",175),
(#"]",#"~",175),
(#"\128",#"\255",175),
(#"\\",#"\\",176)], []), ([(#"\"",#"\"",185),
(#"\\",#"\\",186)], []), ([(#"\t",#"\t",177),
(#"\v",#"\r",177),
(#" ",#" ",177),
(#"\"",#"\"",175),
(#"\\",#"\\",175),
(#"a",#"b",175),
(#"f",#"f",175),
(#"n",#"n",175),
(#"r",#"r",175),
(#"t",#"t",175),
(#"v",#"v",175),
(#"0",#"9",178),
(#"^",#"^",179),
(#"u",#"u",180)], []), ([(#"\t",#"\t",177),
(#"\v",#"\r",177),
(#" ",#" ",177),
(#"\\",#"\\",174)], []), ([(#"0",#"9",184)], []), ([(#"@",#"_",175)], []), ([(#"0",#"9",181),
(#"A",#"F",181),
(#"a",#"f",181)], []), ([(#"0",#"9",182),
(#"A",#"F",182),
(#"a",#"f",182)], []), ([(#"0",#"9",183),
(#"A",#"F",183),
(#"a",#"f",183)], []), ([(#"0",#"9",175),
(#"A",#"F",175),
(#"a",#"f",175)], []), ([(#"0",#"9",175)], []), ([], [62]), ([(#"\t",#"\t",187),
(#"\v",#"\r",187),
(#" ",#" ",187)], []), ([(#"\t",#"\t",187),
(#"\v",#"\r",187),
(#" ",#" ",187),
(#"\\",#"\\",175)], []), ([(#" ",#"!",188),
(#"#",#"[",188),
(#"]",#"~",188),
(#"\128",#"\255",188),
(#"\"",#"\"",189),
(#"\\",#"\\",190)], []), ([], [61]), ([(#"\t",#"\t",191),
(#"\v",#"\r",191),
(#" ",#" ",191),
(#"\"",#"\"",188),
(#"\\",#"\\",188),
(#"a",#"b",188),
(#"f",#"f",188),
(#"n",#"n",188),
(#"r",#"r",188),
(#"t",#"t",188),
(#"v",#"v",188),
(#"0",#"9",192),
(#"^",#"^",193),
(#"u",#"u",194)], []), ([(#"\t",#"\t",191),
(#"\v",#"\r",191),
(#" ",#" ",191),
(#"\\",#"\\",188)], []), ([(#"0",#"9",198)], []), ([(#"@",#"_",188)], []), ([(#"0",#"9",195),
(#"A",#"F",195),
(#"a",#"f",195)], []), ([(#"0",#"9",196),
(#"A",#"F",196),
(#"a",#"f",196)], []), ([(#"0",#"9",197),
(#"A",#"F",197),
(#"a",#"f",197)], []), ([(#"0",#"9",188),
(#"A",#"F",188),
(#"a",#"f",188)], []), ([(#"0",#"9",188)], []), ([(#"\t",#"\t",199),
(#"\v",#"\r",199),
(#" ",#" ",199)], [0])]

    fun mk yyins = let
        (* current start state *)
        val yyss = ref INITIAL
    fun YYBEGIN ss = (yyss := ss)
    (* current input stream *)
        val yystrm = ref yyins
    (* get one char of input *)
    val yygetc = yyInput.getc
    (* create yytext *)
    fun yymktext(strm) = yyInput.subtract (strm, !yystrm)
        open UserDeclarations
        fun lex 
(yyarg as ()) = let 
     fun continue() = let
            val yylastwasn = yyInput.lastWasNL (!yystrm)
            fun yystuck (yyNO_MATCH) = raise Fail "stuck state"
          | yystuck (yyMATCH (strm, action, old)) = 
          action (strm, old)
        val yypos = yyInput.getpos (!yystrm)
        val yygetlineNo = yyInput.getlineNo
        fun yyactsToMatches (strm, [],      oldMatches) = oldMatches
          | yyactsToMatches (strm, act::acts, oldMatches) = 
          yyMATCH (strm, act, yyactsToMatches (strm, acts, oldMatches))
        fun yygo actTable = 
        (fn (~1, _, oldMatches) => yystuck oldMatches
          | (curState, strm, oldMatches) => let
              val (transitions, finals') = Vector.sub (yytable, curState)
              val finals = map (fn i => Vector.sub (actTable, i)) finals'
              fun tryfinal() = 
                    yystuck (yyactsToMatches (strm, finals, oldMatches))
              fun find (c, []) = NONE
            | find (c, (c1, c2, s)::ts) = 
                    if c1 <= c andalso c <= c2 then SOME s
                else find (c, ts)
              in case yygetc strm
              of SOME(c, strm') => 
                   (case find (c, transitions)
                 of NONE => tryfinal()
                  | SOME n => 
                      yygo actTable
                    (n, strm', 
                     yyactsToMatches (strm, finals, oldMatches)))
               | NONE => tryfinal()
              end)
        in 
let
fun yyAction0 (strm, lastMatch : yymatch) = (yystrm := strm; ( continue() ))
fun yyAction1 (strm, lastMatch : yymatch) = (yystrm := strm;
      ( newline yypos; continue() ))
fun yyAction2 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(BANG,      yypos, yytext) )
      end
fun yyAction3 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(PLUS,      yypos, yytext) )
      end
fun yyAction4 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(CAT,       yypos, yytext) )
      end
fun yyAction5 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(MINUS,     yypos, yytext) )
      end
fun yyAction6 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(HASH,      yypos, yytext) )
      end
fun yyAction7 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(LPAR,      yypos, yytext) )
      end
fun yyAction8 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(RPAR,      yypos, yytext) )
      end
fun yyAction9 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(COMMA,     yypos, yytext) )
      end
fun yyAction10 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(ARROW,     yypos, yytext) )
      end
fun yyAction11 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(DOT,       yypos, yytext) )
      end
fun yyAction12 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(COLON,     yypos, yytext) )
      end
fun yyAction13 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(SEAL,      yypos, yytext) )
      end
fun yyAction14 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(SEMICOLON, yypos, yytext) )
      end
fun yyAction15 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(LESS,      yypos, yytext) )
      end
fun yyAction16 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(EQUALS,    yypos, yytext) )
      end
fun yyAction17 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(ISEQUAL,   yypos, yytext) )
      end
fun yyAction18 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(DARROW,    yypos, yytext) )
      end
fun yyAction19 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(LBRACK,    yypos, yytext) )
      end
fun yyAction20 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(RBRACK,    yypos, yytext) )
      end
fun yyAction21 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(UNDERBAR,  yypos, yytext) )
      end
fun yyAction22 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(LBRACE,    yypos, yytext) )
      end
fun yyAction23 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(BAR,       yypos, yytext) )
      end
fun yyAction24 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(RBRACE,    yypos, yytext) )
      end
fun yyAction25 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(BOOL,      yypos, yytext) )
      end
fun yyAction26 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(DATA,      yypos, yytext) )
      end
fun yyAction27 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(DO,        yypos, yytext) )
      end
fun yyAction28 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(ELSE,      yypos, yytext) )
      end
fun yyAction29 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(END,       yypos, yytext) )
      end
fun yyAction30 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(EXPORT,    yypos, yytext) )
      end
fun yyAction31 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(FN,        yypos, yytext) )
      end
fun yyAction32 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(FALSE,     yypos, yytext) )
      end
fun yyAction33 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(FORALL,    yypos, yytext) )
      end
fun yyAction34 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(FUN,       yypos, yytext) )
      end
fun yyAction35 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(IF,        yypos, yytext) )
      end
fun yyAction36 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(IMPORT,    yypos, yytext) )
      end
fun yyAction37 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(IN,        yypos, yytext) )
      end
fun yyAction38 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(INT,       yypos, yytext) )
      end
fun yyAction39 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(LET,       yypos, yytext) )
      end
fun yyAction40 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(LINK,      yypos, yytext) )
      end
fun yyAction41 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(MODULE,    yypos, yytext) )
      end
fun yyAction42 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(NEW,       yypos, yytext) )
      end
fun yyAction43 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(OF,        yypos, yytext) )
      end
fun yyAction44 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(OPEN,      yypos, yytext) )
      end
fun yyAction45 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(OUT,       yypos, yytext) )
      end
fun yyAction46 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(PRINT,     yypos, yytext) )
      end
fun yyAction47 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(REC,       yypos, yytext) )
      end
fun yyAction48 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(SEALS,     yypos, yytext) )
      end
fun yyAction49 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(SIGNATURE, yypos, yytext) )
      end
fun yyAction50 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(STRING,    yypos, yytext) )
      end
fun yyAction51 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(THEN,      yypos, yytext) )
      end
fun yyAction52 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(TRUE,      yypos, yytext) )
      end
fun yyAction53 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(TYPE,      yypos, yytext) )
      end
fun yyAction54 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(UNIT,      yypos, yytext) )
      end
fun yyAction55 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(VAL,       yypos, yytext) )
      end
fun yyAction56 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(WHERE,     yypos, yytext) )
      end
fun yyAction57 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( token(WITH,      yypos, yytext) )
      end
fun yyAction58 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( tokenOf(NUM,     toInt,    yypos, yytext) )
      end
fun yyAction59 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( tokenOf(HEXNUM,  toHexInt, yypos, yytext) )
      end
fun yyAction60 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( tokenOf(REAL,    toReal,   yypos, yytext) )
      end
fun yyAction61 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( tokenOf(TEXT,    toString, yypos, yytext) )
      end
fun yyAction62 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( tokenOf(CHAR,    toChar,   yypos, yytext) )
      end
fun yyAction63 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( tokenOf(TYPVAR,  toId,     yypos, yytext) )
      end
fun yyAction64 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( tokenOf(ALPHA,   toId,     yypos, yytext) )
      end
fun yyAction65 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( tokenOf(SYMBOL,  toId,     yypos, yytext) )
      end
fun yyAction66 (strm, lastMatch : yymatch) = (yystrm := strm;
      ( YYBEGIN LCOMMENT; continue() ))
fun yyAction67 (strm, lastMatch : yymatch) = (yystrm := strm;
      ( nest yypos; YYBEGIN COMMENT; continue() ))
fun yyAction68 (strm, lastMatch : yymatch) = (yystrm := strm; ( continue() ))
fun yyAction69 (strm, lastMatch : yymatch) = (yystrm := strm;
      ( YYBEGIN (if !nesting = 0 then INITIAL else COMMENT);
                          newline yypos; continue() ))
fun yyAction70 (strm, lastMatch : yymatch) = (yystrm := strm;
      ( YYBEGIN LCOMMENT; continue() ))
fun yyAction71 (strm, lastMatch : yymatch) = (yystrm := strm;
      ( nest yypos; continue() ))
fun yyAction72 (strm, lastMatch : yymatch) = (yystrm := strm;
      ( unnest();
                          if !nesting = 0 then YYBEGIN INITIAL else ();
                          continue() ))
fun yyAction73 (strm, lastMatch : yymatch) = (yystrm := strm; ( continue() ))
fun yyAction74 (strm, lastMatch : yymatch) = (yystrm := strm;
      ( newline yypos; continue() ))
fun yyAction75 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( error(yypos, yytext, "invalid string") )
      end
fun yyAction76 (strm, lastMatch : yymatch) = let
      val yytext = yymktext(strm)
      in
        yystrm := strm; ( invalid(yypos, yytext) )
      end
val yyactTable = Vector.fromList([yyAction0, yyAction1, yyAction2, yyAction3,
  yyAction4, yyAction5, yyAction6, yyAction7, yyAction8, yyAction9, yyAction10,
  yyAction11, yyAction12, yyAction13, yyAction14, yyAction15, yyAction16,
  yyAction17, yyAction18, yyAction19, yyAction20, yyAction21, yyAction22,
  yyAction23, yyAction24, yyAction25, yyAction26, yyAction27, yyAction28,
  yyAction29, yyAction30, yyAction31, yyAction32, yyAction33, yyAction34,
  yyAction35, yyAction36, yyAction37, yyAction38, yyAction39, yyAction40,
  yyAction41, yyAction42, yyAction43, yyAction44, yyAction45, yyAction46,
  yyAction47, yyAction48, yyAction49, yyAction50, yyAction51, yyAction52,
  yyAction53, yyAction54, yyAction55, yyAction56, yyAction57, yyAction58,
  yyAction59, yyAction60, yyAction61, yyAction62, yyAction63, yyAction64,
  yyAction65, yyAction66, yyAction67, yyAction68, yyAction69, yyAction70,
  yyAction71, yyAction72, yyAction73, yyAction74, yyAction75, yyAction76])
in
  if yyInput.eof(!(yystrm))
    then UserDeclarations.eof(yyarg)
    else (case (!(yyss))
       of COMMENT => yygo yyactTable (0, !(yystrm), yyNO_MATCH)
        | LCOMMENT => yygo yyactTable (1, !(yystrm), yyNO_MATCH)
        | INITIAL => yygo yyactTable (2, !(yystrm), yyNO_MATCH)
      (* end case *))
end
            end
      in 
            continue()       
        handle IO.Io{cause, ...} => raise cause
          end
        in 
          lex 
        end
    in
    fun makeLexer yyinputN = mk (yyInput.mkStream yyinputN)
    fun makeLexer' ins = mk (yyInput.mkStream ins)
    end

  end
